// Designed for ATTiny85 and LiPo+BMS
// or       for ATTiny85V and 2xAA batteries
//
// Activity LED on PB0 (pin6)
// 940nm IR LED on PB1 (pin5)


/*
  IR Remote Control
  129 Left
  130 Forward
  131 Right
  132 Spot
  133 Max
  134 Small
  135 Medium
  136 Large / Clean
  137 Stop
  138 Power
  139 Arc Left
  140 Arc Right
  141 Stop
  Scheduling Remote
  142 Download
  143 Seek Dock
  Roomba Discovery Driveon Charger
  240 Reserved
  248 Red Buoy
  244 Green Buoy
  242 Force Field
  252 Red Buoy and Green Buoy
  250 Red Buoy and Force Field
  246 Green Buoy and Force Field
  254 Red Buoy, Green Buoy and Force
  Roomba 500 Drive-on Charger
  160 Reserved
  161 Force Field
  164 Green Buoy
  165 Green Buoy and Force Field
  168 Red Buoy
  169 Red Buoy and Force Field
  172 Red Buoy and Green Buoy
  173 Red Buoy, Green Buoy and Force Field
  Roomba 500 Virtual Wall
  162 Virtual Wall
  Roomba 500 Virtual Wall Lighthouse ###### (FIXME: not yet supported here)
  0LLLL0BB
  LLLL = Virtual Wall Lighthouse ID
  (assigned automatically by Roomba
  560 and 570 robots)
  1-10: Valid ID
  11: Unbound
  12-15: Reserved
  BB = Which Beam
  00 = Fence
  01 = Force Field
  10 = Green Buoy
  11 = Red Buoy
*/

#include <avr/wdt.h>
#include <avr/power.h>
#include <avr/sleep.h>

#ifdef F_CPU
#  define SYSCLOCK  F_CPU     // main Arduino clock
#else
# define SYSCLOCK  8000000  // main Arduino clock
#endif

#define LED_PIN              PB0
// pinMode(LED_PIN, OUTPUT); digitalWrite(LED_PIN, LOW);
#define LED_SETUP ({ \
    DDRB |= (1 << LED_PIN); \
    PORTB &= ~(1 << LED_PIN); \
  })
#define LED_ON ({PORTB |= (1 << LED_PIN);})   // digitalWrite(LED_PIN, HIGH);
#define LED_OFF ({PORTB &= ~(1 << LED_PIN);}) // digitalWrite(LED_PIN, LOW);

#define IR_PIN              PB1
#define IR_SETUP ({ \
    DDRB |= (1 << IR_PIN); \
    PORTB &= ~(1 << IR_PIN); \
  })

// PWM Carrier
// FastPWM-Compare mode, no prescale (to support 1MHz internal clock)
// Disable interruptions (Arduino libs have ISR for delay and millis)
#define PWM_SETUP(val) ({ \
    const uint8_t pwmval = SYSCLOCK / val / 1000; \
    TCCR0A = 1<<WGM00 | 1<<WGM01; \
    TCCR0B = 1<<WGM02 | 1<<CS00; \
    TIMSK &= 0b10000001; \
    OCR0A  = pwmval;\
    OCR0B  = pwmval/2; \
  })
#define PWM_ON  ({TCNT0=0; TCCR0A |= _BV(COM0B1);})
#define PWM_OFF (TCCR0A &= ~(_BV(COM0B1)))


long readVcc();
void roomba_send(char code);
void wdtSleep (unsigned int ms);
void delay_ten_us(unsigned long us);
void custom_delay_usec(unsigned long uSecs);

unsigned int code = 162; // Virtual Wall
unsigned int iter = 0;
unsigned int timeout = 0; // approx 154 minutes before overflow sleep
long bat = 0;

void setup() {
  noInterrupts();  // no need to interrupt when not sleeping
  LED_SETUP;
  IR_SETUP;
  PWM_SETUP(38);   // 38kHz carrier
  power_timer1_disable();
  power_adc_disable();
  wdt_disable(); // In case it was rebooted during sleep, for any reason, without power cycle, wdt is still enabled
}

void loop() {
  iter++;
  timeout++;

  if      (iter == 1) bat = readVcc();
  else if (iter == 2) LED_ON;
  else if (iter == 3) LED_OFF;
  else if ((iter == 4) && (bat < 3000)) LED_ON;   // Thresholds in mV for 1S LiPo + BMS
  else if ((iter == 5) && (bat < 3000)) LED_OFF;
  else if ((iter == 6) && (bat < 2800)) LED_ON;
  else if ((iter == 7) && (bat < 2800)) LED_OFF;
  else if (iter == 120) iter = 0;                 // New battCheck every 15s approx

  roomba_send(code); // 4ms
  wdtSleep(100);     // 100 ms

  // Extra 50ms break every 3 burst to have a 150ms break
  if ((iter % 3) == 0)
    wdtSleep(50);    // 50ms

  if (timeout == 0) {
    set_sleep_mode (SLEEP_MODE_PWR_DOWN);
    power_all_disable();
    noInterrupts();
    sleep_mode();
  }
}


/*******************************************
   Internal battery ADC measuring
   from sensebender micro sketch
 *******************************************/
long readVcc() {
  power_adc_enable();
  // Read 1.1V reference against AVcc
  // set the reference to Vcc and the measurement to the internal 1.1V reference
#if defined(__AVR_ATmega32U4__) || defined(__AVR_ATmega1280__) || defined(__AVR_ATmega2560__)
  ADMUX = _BV(REFS0) | _BV(MUX4) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
#elif defined (__AVR_ATtiny24__) || defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
  ADMUX = _BV(MUX5) | _BV(MUX0);
#elif defined (__AVR_ATtiny25__) || defined(__AVR_ATtiny45__) || defined(__AVR_ATtiny85__)
  ADMUX = _BV(MUX3) | _BV(MUX2);
#else
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
#endif
  custom_delay_usec(2000); // 2ms enough

  ADCSRA |= _BV(ADSC); // Start conversion
  while (bit_is_set(ADCSRA, ADSC)); // measuring

  uint8_t low  = ADCL; // must read ADCL first - it then locks ADCH
  uint8_t high = ADCH; // unlocks both

  long result = (high << 8) | low;

  result = 1125300L / result; // Calculate Vcc (in mV); 1125300 = 1.1*1023*1000
  power_adc_disable();
  return result; // Vcc in millivolts
}

void roomba_send(char code) {
  for (char i = 7; i >= 0; i--) {
    if (code & (1 << i)) {
      PWM_ON;
      custom_delay_usec(3000);
      PWM_OFF;
      custom_delay_usec(1000);
    } else {
      PWM_ON;
      custom_delay_usec(1000);
      PWM_OFF;
      custom_delay_usec(3000);
    }
  }
}

// watchdog interrupt
ISR (WDT_vect) {
  wdt_disable();
}

void wdtSleep (unsigned int ms) {
  power_timer0_disable();
  set_sleep_mode (SLEEP_MODE_PWR_DOWN);
  ms = ms / 16;
  // 0: 16 ms, 1: 32 ms, 2: 64 ms, 3: 128 ms, 4: 256 ms,
  // 5: 512 ms, 6: 1024 ms, 7: 2048 ms, 8: 4096 ms, 9: 8192 ms,
  unsigned char wdp = 0;

  while ((ms) && (wdp <= 9)) {
    if (ms & 1) {
      MCUSR = 0;                              // clear various "reset" flags
      //      noInterrupts();                         // Timed sequence follow
      WDTCR = (1 << WDCE | 1 << WDE);         // watchdog change enable
      WDTCR |= (1 << WDIE | 0 << WDE | wdp) ; // enable wdt interrupt
      //      interrupts();      // End of timed sequence

      // turn off brown‐out enable in software
      MCUCR = bit (BODS) | bit (BODSE);
      MCUCR = bit (BODS);

      wdt_reset();
      interrupts();      // waiting for watchdog interrupt to wakeup
      sleep_mode ();
      noInterrupts();    // no need for interruptions in this sketch
      sleep_disable();
    }
    ms = ms >> 1;
    wdp = wdp + 1;
  }
  power_timer0_enable();
}

// Busy loop to manage very small delays (<16ms) that
// can not be managed by wdtSleep timer
// Inspired by TV B Gone project
// TODO: use Timer0 (38kHz), SLEEP_MODE_IDLE and interrupt
//       if not too long for PWM signal generation
#define DELAY_CNT SYSCLOCK/1000000
#define NOP __asm__ __volatile__ ("nop")

void delay_ten_us(unsigned long us) {
  uint8_t timer;
  while (us != 0) {
    for (timer = 0; timer <= DELAY_CNT; timer++) {
      NOP;
      NOP;
    }
    us--;
  }
}

void custom_delay_usec(unsigned long uSecs) {
  delay_ten_us(uSecs / 10);
}
